import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.Arrays; 
import java.util.IntSummaryStatistics; 
import peasy.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class checkoutBarChart2 extends PApplet {




/*************************************************************************************

--  Data Visualization - 3D Visualization
    Tested in Processing 3

    Author: Alexander (Sandy) Schoettler
    Supervisor: George Legrady

    Purpose: Show a bar chart of checkout information for a given item type (book or video),
        grouped by day of the week & week of the year.
        Also compare this data to similar data from previous years.

    Usage: 1. A mouse left-drag will rotate the camera around the subject.
                2. A right drag will zoom in and out.
                3. A middle-drag (command-left-drag on mac) will pan.
                4. A double-click restores the camera to its original position.
                5. The shift key constrains rotation and panning to one axis or the other.

--  NOTES

-- Dates in 2016:
        Jan 1 - Jan 3 : data added to week of Jan 4
        Jan 4: start of first full week in data
        MLK Jr  Day         M   1/18    w3
        President's Day     M   2/15    w7

        Memorial day        M   5/30    w22
        Independence Day    M   7/4     w27

        Labor Day           M   9/5     w36 ** large spike next day
        Election Day        T   11/8    w45
        Veterans Day        F   11/11   w45

        Summer start        M   6/20    w25 ** start of more activity
        Summer end          R   9/22    w38

-- Trends in 2016:
    -   Many holidays fall on Monday. Often the following day has spike in activity
    -   Start of May in week 18 Sunday activity decreased until start of summer in late June
    -   Despite Sunday being one of the less popular days of the week, its popularity increased since 2012.
    -   In summer months, there's more traffic overall, but especially on earlier days of the week.
    -   In fact Saturday activity decreased during that period from its otherwise higher activity
    -   The early-week summer days in 2012 had similar activity to 2016 (as seen by color)

-- Other Trends:
    -   In 2013 there is a seasonal increase in summer activity early in the week, complemented by
        an increase in winter activity on Saturdays.
    -   DVD ratios down (blue) from 2013 to 2015


-- Color Schemes --
Available Colors: plasma, BuRd, magma, viridis

-- Key Commands --
    Key     Description     scheme  pos
    ------------------------------------
    0       fixed color     none    n/a
    1       Differences     plasma  0
    2       Differences     RB      1
    3       Ratios          RB      1
    4       Actual Value    magma   2
*************************************************************************************/


PeasyCam cam;
float[] rotations = new float[3];


// Constants
final int XMIN = -20;     // WEEKS : M T W R F S S
final int XMAX = 100;
final int YMIN = -50;     // Week Index : 0 to 52
final int YMAX = 400;
final int ZMIN = -90;
final int ZMAX = 180;   // Height

final int BOOK_TYPE = 0;
final int DVD_TYPE = 1;
final int NUM_ITEMTYPES = 2;

final int NUM_YEARS = 9;
final int NUM_WEEKS = 53;
final int TSLICES_PER_WEEK = 7 * 24 * 4;
final int DAYS_PER_WEEK = 7;
final int TSLICES_PER_DAY = 24 * 4;



/***** -- Variables to choose data -- *****/
//  year2 = current year = ((actual year) - (2010)), so (year2 = {0,1,2,3,4,5,6,7,8})
//  year1 = other year for comparison
int type = BOOK_TYPE;
int year2 = 6; // =6 for 2016 data
int year1 = 2; // =2 for 2012 data


// Data Structures
Table table;
Table table_books_1;
Table table_books_2;
Table table_dvd_1;
Table table_dvd_2;

int[][][][] checkouts = new int[NUM_ITEMTYPES][NUM_YEARS][NUM_WEEKS][TSLICES_PER_WEEK];
int[][] annualMaxValues = new int[NUM_ITEMTYPES][NUM_YEARS];
float[][][][] daily_checkouts = new float[NUM_ITEMTYPES][NUM_YEARS][NUM_WEEKS][DAYS_PER_WEEK];
float[][] annualDailyMaxValues = new float[NUM_ITEMTYPES][NUM_YEARS];



// Graphics Variables
PImage colorBar;
String[] colorPath = {"plasma.png", "BuRd.png", "magma.png", "viridis.png"};
boolean colorScheme = false;
int colorSchemeID = 0;
float diff_scale = 60;
float ratio_scale = 0.8f;

ArrayList<MyPolygon> polygons = new ArrayList<MyPolygon>();
float RIBBON_TOP_ALPHA = 254;
float RIBBON_DIAG_ALPHA = 80;

float RIBBON_PORTION_X = 0.8f;
float RIBBON_PORTION_Y = 0.7f;

float BOX_THICKNESS = 2;


public int computeColor(int week, int day) {
    if (colorSchemeID == 1 || colorSchemeID == 2) {
        float num1 = daily_checkouts[type][year1][week][day];
        float num2 = daily_checkouts[type][year2][week][day];
        float val = (num2 - num1);
        int c = colorBar.get((int)map(clamp_rangef(-diff_scale, diff_scale, val), -diff_scale, diff_scale, 2, colorBar.width-2), colorBar.height/2);
        return c;
    }
    if (colorSchemeID == 3) {
        float num1 = daily_checkouts[type][year1][week][day];
        float num2 = daily_checkouts[type][year2][week][day];
        float val = (num2 - num1) / (num1 + 1);
        int c = colorBar.get((int)map(clamp_rangef(-ratio_scale, ratio_scale, val), -ratio_scale, ratio_scale, 2, colorBar.width-2), colorBar.height/2);
        return c;
    }
    if (colorSchemeID == 4) {
        float val = daily_checkouts[type][year2][week][day];
        int c = colorBar.get((int)map(val, 0, annualDailyMaxValues[type][year2], 2, colorBar.width-2), colorBar.height/2);
        return c;
    }
    else {
        return 0xffE5D2FF;
    }
}
public int clamp_range(int min, int max, int val) {
    if (val < min)
        return min;
    else if (val > max)
        return max;
    else
        return val;
}
public float clamp_rangef(float min, float max, float val) {
    if (val < min)
        return min;
    else if (val > max)
        return max;
    else
        return val;
}

public void setup() {
    
    
    textMode(SHAPE);


    /* --- CAMERA --- */
    cam = new PeasyCam(this, (XMIN+XMAX)/2, (YMIN-YMAX)/2, (ZMIN+ZMAX)/2, 200);
    float cameraZ = (height/2.0f) / tan(PI*60.0f/360.0f);
    perspective(PI/2, width/height, 0.001f, cameraZ*10.0f);
    cam.setWheelScale(0.5f);
    println("Camera Position:");
    println(cam.getPosition());
    println("Camera Target: ");
    println(cam.getLookAt());

    println("distance: " + cam.getDistance());

    /* --- CSV DATA --- */
    table = loadTable("data.tsv", "header, tsv");

    // ** checkouts
    for (TableRow row : table.rows()) {
        int type = row.getInt("itemType");
        int yearNum = row.getInt("yearNum");
        int weekNum = clamp_range(1, 52, row.getInt("weekNum")) - 1; // NOTE changed to 52 to get data for last week.
        int tslice = row.getInt("tslice");
        int qty = row.getInt("qty");
        checkouts[type][yearNum][weekNum][tslice] = qty;
    }

    // daily_checkouts = daily averages from checkouts
    for (int t=0; t < NUM_ITEMTYPES; t++) {
        for (int y=0; y < NUM_YEARS; y++) {
            for (int w=0; w < NUM_WEEKS; w++) {
                for (int day = 0; day < DAYS_PER_WEEK; day++) {
                    float sum = 0;
                    for (int dayslice=0; dayslice < TSLICES_PER_DAY; dayslice++) {
                        int tslice = day * TSLICES_PER_DAY + dayslice;
                        sum += checkouts[t][y][w][tslice];
                    }
                    daily_checkouts[t][y][w][day] = sum / TSLICES_PER_DAY;
                }
            }
        }
    }

    // Type/Year max for daily_checkouts
    for(int t = 0; t < NUM_ITEMTYPES; t++) {
        for (int y=0; y < NUM_YEARS; y++) {
            float max = daily_checkouts[t][y][0][0];
            for (float[] arr1 : daily_checkouts[t][y]) {
                for (float val : arr1) {
                    if (val > max) {
                        max = val;
                    }
                }
            }
            annualDailyMaxValues[t][y] = max;
        }
    }

    /* --- COLORS --- */
    initColorScheme();

    setup_polygons();

}
int frame_count = 0;
public void draw() {
    background(30);
    draw_labels();
    draw_boxes();
    draw_polygons();
}

public void setup_polygons() {
    pushMatrix();
    translate(XMIN, YMIN, ZMIN);
    scale(1, -1, 1);
    scale((1.0f * (XMAX-XMIN))/TSLICES_PER_WEEK, (1.0f * (YMAX-YMIN))/NUM_WEEKS, 1);

    for (int day=0; day < DAYS_PER_WEEK; day++) {
        for (int w=0; w < NUM_WEEKS-1; w++) {
            float boxdimX = TSLICES_PER_DAY;
            float box_x0 = day * boxdimX;

            float boxdimY = 1;
            float box_y0 = w * boxdimY;
            float box_y2 = (w+1) * boxdimY;

            float box_z0 = 0;
            float box_z1 = daily_checkouts[type][year2][w][day];
            float boxdimZ1 = box_z1 - box_z0;
            float box_z2 = daily_checkouts[type][year2][w+1][day];
            float boxdimZ2 = box_z2 - box_z0;

            PShape flattop = createShape();
            flattop.beginShape(QUAD);
            flattop.fill(color(computeColor(w, day), RIBBON_TOP_ALPHA));
            flattop.fill(color(255,0,0));
            flattop.strokeWeight(1);
            flattop.stroke(255);
            flattop.vertex(box_x0 + boxdimX * (1 - RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 - RIBBON_PORTION_Y)/2, 0);
            flattop.vertex(box_x0 + boxdimX * (1 + RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 - RIBBON_PORTION_Y)/2, 0);
            flattop.vertex(box_x0 + boxdimX * (1 + RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 + RIBBON_PORTION_Y)/2, 0);
            flattop.vertex(box_x0 + boxdimX * (1 - RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 + RIBBON_PORTION_Y)/2, 0);
            flattop.endShape(CLOSE);
            polygons.add(new MyPolygon(flattop, w, day, true));

            if (w < NUM_WEEKS - 2) {
                PShape ramp = createShape();
                ramp.beginShape(QUAD);
                ramp.fill(color(computeColor(w, day), RIBBON_DIAG_ALPHA));
                ramp.fill(color(0,0,255));
                ramp.strokeWeight(1);
                ramp.stroke(255);
                ramp.vertex(box_x0 + boxdimX * (1 - RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 + RIBBON_PORTION_Y)/2, box_z1);
                ramp.vertex(box_x0 + boxdimX * (1 + RIBBON_PORTION_X)/2, box_y0 + boxdimY * (1 + RIBBON_PORTION_Y)/2, box_z1);
                ramp.vertex(box_x0 + boxdimX * (1 + RIBBON_PORTION_X)/2, box_y2 + boxdimY * (1 - RIBBON_PORTION_Y)/2, box_z2);
                ramp.vertex(box_x0 + boxdimX * (1 - RIBBON_PORTION_X)/2, box_y2 + boxdimY * (1 - RIBBON_PORTION_Y)/2, box_z2);
                ramp.endShape(CLOSE);
                polygons.add(new MyPolygon(ramp, w, day, false));
            }
        }
    }
    popMatrix();
}

public void draw_labels() {
    fill(255);
    pushMatrix();
    translate(XMIN, YMIN, ZMIN); // absolute corner
    scale(1, -1, 1);

    // X AXIS -- MTWRFSS
    String weekdays = "MTWRFSS";
    textSize(12);
    pushMatrix();
        textAlign(CENTER, TOP);
        for (int w=0; w < DAYS_PER_WEEK; w++) {
            pushMatrix();
                translate((w + 0.5f) * (XMAX-XMIN)/7, 0, 0);
                rotateX(PI);
                text(weekdays.charAt(w), 0, 0, 0);
            popMatrix();
        }
    popMatrix();
    pushMatrix();
        textSize(18);
        translate((XMAX-XMIN)/2, -5 * (YMAX-YMIN)/NUM_WEEKS, 0);
        textAlign(CENTER, CENTER);
        rotateX(PI);
        text("Weekday", 0, 0, 0);
    popMatrix();
    //

    // Y AXIS -- Week Numbers 1-52
    pushMatrix();
    textSize(7);
    noStroke();
    fill(0xffFFFFFF);
    for (int w=0; w < NUM_WEEKS-1; w++) {
        pushMatrix();
            translate((XMAX-XMIN) * (7.5f/7.0f), (w+0.5f) * (YMAX-YMIN)/NUM_WEEKS, 0);
            // rotateY(PI);
            rotateX(PI);
            // sphere(3);
            textAlign(CENTER, CENTER);
            text(Integer.toString(w+1), 0, 0, 0);
        popMatrix();
    }

    // draw_annotation(36, "Labor Day");

    popMatrix();
    //

    popMatrix();
}

public void draw_annotation(int w, String label) {
    pushMatrix();
    translate((XMAX-XMIN) * (9/7.0f), (w+0.5f) * (YMAX-YMIN)/NUM_WEEKS, 0);
    rotateX(PI);
    textAlign(LEFT, CENTER);
    text(label, 0, 0, 0);
    popMatrix();
}

public void draw_boxes() {
    pushMatrix();
    translate(XMIN, YMIN, ZMIN);
    scale(1, -1, 1);
    scale((1.0f * (XMAX-XMIN))/TSLICES_PER_WEEK, (1.0f * (YMAX-YMIN))/NUM_WEEKS, 1);

    // line(x0 + boxdimX/5, y0 + boxdimY/2, 0, x0 + boxdimX/5, y0 + boxdimY/2, z1);
    // line(x0 + 4*boxdimX/5, y0 + boxdimY/2, 0, x0 + 4*boxdimX/5, y0 + boxdimY/2, z1);

    float boxdimX = TSLICES_PER_DAY;
    float boxdimY = 1;
    for (int day=0; day < DAYS_PER_WEEK; day++) {
        for (int w=0; w < NUM_WEEKS-1; w++) {
            float x0 = day * TSLICES_PER_DAY;
            float y0 = w * boxdimY;
            float z0 = 0;
            float z1 = daily_checkouts[type][year2][w][day];
            float boxdimZ = z1 - z0;

            fill(computeColor(w, day), 255);
            strokeWeight(0.8f);
            stroke(128);
            pushMatrix();
            translate(x0 + boxdimX/2, y0 + boxdimY/2, z0 + boxdimZ);
            box(boxdimX * RIBBON_PORTION_X * 0.99f, boxdimY * RIBBON_PORTION_Y * 0.99f, BOX_THICKNESS);
            popMatrix();
        }
    }

    popMatrix();
}

public void draw_polygons() {
    pushMatrix();
    translate(XMIN, YMIN, ZMIN);
    scale(1, -1, 1);
    scale((1.0f * (XMAX-XMIN))/TSLICES_PER_WEEK, (1.0f * (YMAX-YMIN))/NUM_WEEKS, 1);

    for (MyPolygon mypoly : polygons) {
        int theColor = color(255, 0, 0);
        if (mypoly.isTop()) {
            theColor = color(computeColor(mypoly.week, mypoly.day), (int)RIBBON_TOP_ALPHA);
            mypoly.setFill(theColor);
            // shape(mypoly.getShape());
        } else {
            theColor = color(255, (int)RIBBON_DIAG_ALPHA);
            mypoly.setFill(theColor);
            shape(mypoly.getShape());
        }
    }
    popMatrix();
}



public void keyPressed() {
    if (Arrays.asList('0','1','2','3','4').contains(key)) {
        int num = (int)(key - '0');
        if (num == 0) {
            colorScheme = false;
            colorSchemeID = 0;
            return;
        }
        colorScheme = true;
        colorSchemeID = num;
        if (num == 1) {
            colorBar = loadImage(colorPath[0]);
        }
        if (num == 2) {
            colorBar = loadImage(colorPath[1]);
        }
        if (num == 3) {
            colorBar = loadImage(colorPath[1]);
        }
        if (num == 4) {
            colorBar = loadImage(colorPath[3]);
        }
    }
    if(key == 's' || key == 'S'){
        saveFrame();
    }
}

public void initColorScheme() {
    colorScheme = true;
    colorSchemeID = 4;
    colorBar = loadImage(colorPath[3]);
}

public void draw_basis(float sc) {
    shapeMode(CENTER);
    pushMatrix();
    noStroke();
    fill(0xffFFFFFF, 128);
    sphere(2);

    pushMatrix();
    translate(sc, 0, 0);
    fill(0xffFF0000, 128);
    box(2*sc, sc/8, sc/8);
    translate(7*sc, 0, 0);
    sphere(2);
    popMatrix();

    pushMatrix();
    translate(0, sc, 0);
    fill(0xff00FF00, 128);
    box(sc/8, 2*sc, sc/8);
    translate(0, 7*sc, 0);
    sphere(2);
    popMatrix();

    pushMatrix();
    translate(0, 0, sc);
    fill(0xff0000FF, 128);
    box(sc/8, sc/8, 2*sc);
    translate(0, 0, 7*sc);
    sphere(2);
    popMatrix();

    popMatrix();
}

public void ball(float x, float y, float z) {
    pushMatrix();
    translate(x, y, z);
    sphere(4);
    popMatrix();
}
public class MyPolygon {
    public PShape poly;
    public int week;
    public int day;
    boolean isTop = true;

    MyPolygon(PShape poly, int w, int d, boolean isTop) {
        this.poly = poly;
        this.week = w;
        this.day = d;
        this.isTop = isTop;
    }

    public void setFill(int col) {
        poly.setFill(col);
    }

    public PShape getShape() {
        return poly;
    }
    public boolean isTop() {
        return isTop;
    }
    public int getWeek() {
        return week;
    }
    public int getDay() {
        return day;
    }

}
  public void settings() {  size(1600, 1000, P3D);  smooth(3); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "checkoutBarChart2" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
